//
//  RectangularTileMaker.cpp
//  tilemaker
//
//  Created by Attila Zsaki on 2015-10-28.
//  Copyright © 2015 Attila Zsaki. All rights reserved.
//

#include <stdio.h>
#include <iostream>

#include "UtilityFunctions.hpp"
#include "RectangularTileMaker.hpp"

using namespace std;

RectangularTileMaker::RectangularTileMaker()
{
    dTiledAreaWidth=10.0;
    dTiledAreaLength=10.0;
    
    dTileWidth=1.0;
    dTileLength=1.0;
    
    dGroutWidth=0.1;
    
    dTileXOffset=0.0;

    
    tileColorDescriptorGlobal=ColorDescriptor();
    tileColorDescriptorLocal=ColorDescriptor();
    groutColorDescriptor=ColorDescriptor();
    
    iTextureWidth=1024;
    iTextureHeight=1024;
    
    bRandomTileOrientation=false;
    bRandomTileNoiseStartPoint=false;
    
    bUseTileXOffset=false;
    
    bGenerateBumpMap=false;
    
    dMaxGroutBumpHeight=0.6;
    dMaxTileBumpHeight=1.0;
    
    bTileBumpNoiseAlternate=false;
    tileColorDescriptorBump=ColorDescriptor();
    
    bGenerateDiscolorationMap=false;
    tileColorDescriptorDiscoloration=ColorDescriptor();
    
    dDiscolorationThresholdLower=0.0;
    dDiscolorationThresholdUpper=1.0;
    dDiscolorationThresholdPercentReduction=0.0;
    
    bBlurFinalImage=false;

}

RectangularTileMaker::~RectangularTileMaker()
{
    
}

void RectangularTileMaker::GenerateTexture()
{
    
    // 1. create the texture and bump image
    Image textureImage(Geometry(iTextureWidth,iTextureHeight),Color(MaxRGB,MaxRGB,MaxRGB,0));
    Image bumpImage(Geometry(iTextureWidth,iTextureHeight),Color(MaxRGB,MaxRGB,MaxRGB,0));
    
    // 2. render the grout color filling the whole texture
    RenderGrout(&textureImage,&bumpImage);
    
    textureImage.blur(2.0,0.5);
    
    // 3. render the tiles
    RenderTiles(&textureImage,&bumpImage);
    
    // 4. render discoloration map if needed
    if (bGenerateDiscolorationMap) {
        RenderDiscoloration(&textureImage);
    }
    
    // blur if neeeded
    if (bBlurFinalImage) {
        textureImage.blur(2.0,0.5);
    }
    
    //
    textureImage.write(outputFileName);
    
    if (bGenerateBumpMap) {
        bumpImage.write(outputFileName+"_bump.tif");
    }
    
}

void RectangularTileMaker::TileToTextureSpace(double x, double y, int* i, int* j)
{
    double dx=x/dTiledAreaWidth;
    double dy=y/dTiledAreaLength;
    
    *i=(int)(iTextureWidth*dx);
    *j=(int)(iTextureHeight*dy);
}

void RectangularTileMaker::TextureToTileSpace(int i, int j, double* x, double* y)
{
    double dx=i/(double)iTextureWidth;
    double dy=j/(double)iTextureHeight;
    
    *x=dTiledAreaWidth*dx;
    *y=dTiledAreaLength*dy;
}

void RectangularTileMaker::RenderGrout(Image* textureImage,Image* bumpImage)
{
    
    for (int j=0;j<iTextureHeight;++j) {
        for (int i=0;i<iTextureWidth;++i) {
            
            // tile space coordinates
            double x,y;
            
            TextureToTileSpace(i,j,&x,&y);
            
            // get color
            double noiseValue;
            ColorRGBA color=groutColorDescriptor.GetColor(x,y,&noiseValue);
            
            textureImage->pixelColor(i,j,Color(color.dR*MaxRGB,color.dG*MaxRGB,color.dB*MaxRGB,0));
            
            noiseValue*=dMaxGroutBumpHeight;
            
            // set bump
            bumpImage->pixelColor(i,j,Color(noiseValue*MaxRGB,noiseValue*MaxRGB,noiseValue*MaxRGB,0));
        }
    }
}

void RectangularTileMaker::RenderTiles(Image* textureImage,Image* bumpImage)
{
    
    int numFullTilesX;
    int numFullTilesY;
    
    numFullTilesX=(int)(dTiledAreaWidth/(dTileWidth+dGroutWidth))+1;
    numFullTilesY=(int)(dTiledAreaLength/(dTileLength+dGroutWidth))+1;
    
    

    for (int j=0;j<numFullTilesY;++j) {
        
        int numFullTilesXToUse=numFullTilesX;
        
        if (bUseTileXOffset) {
            if (j%2==1) {
                numFullTilesXToUse+=1;
            }
        }
        
        for (int i=0;i<numFullTilesXToUse;++i) {
            
            // tile lower left coordinates
            double tileStartX;
            double tileStartY;
            int tileStartI;
            int tileStartJ;
            
            // tile upper right coordinates
            double tileEndX;
            double tileEndY;
            int tileEndI;
            int tileEndJ;
            
            tileStartX=i*(dTileWidth+dGroutWidth);
            tileStartY=j*(dTileLength+dGroutWidth);
            tileEndX=tileStartX+(dTileWidth);
            tileEndY=tileStartY+(dTileLength);
            
            // tile offsets, if used
            if (bUseTileXOffset) {
                if (j%2==1) {
                    tileStartX-=dTileXOffset;
                    tileEndX-=dTileXOffset;
                }
            }
            
            //cout << tileStartX << " " << tileStartY << endl;
            //cout << tileEndX << " " << tileEndY << endl;
            
            TileToTextureSpace(tileStartX,tileStartY,&tileStartI,&tileStartJ);
            TileToTextureSpace(tileEndX,tileEndY,&tileEndI,&tileEndJ);
            
            //cout << tileStartI << " " << tileStartJ << endl;
            //cout << tileEndI << " " << tileEndJ << endl;
            
            
            // find the color variation in the tile on the global scale
            double tileCenterX=((tileStartX+tileEndX)*0.5)/dTiledAreaWidth;
            double tileCenterY=((tileStartY+tileEndY)*0.5)/dTiledAreaLength;
            double noiseValue;
            ColorRGBA colorGlobal=tileColorDescriptorGlobal.GetColor(tileCenterX,tileCenterY,&noiseValue);
            
            // see if random texture layout is desired
            bool flipOrientation=false;
            
            if (bRandomTileOrientation) {
                UtilityFunctions utility;
                
                if (utility.RandomNumberInRange(0.0,1.0)>0.5) {
                    flipOrientation=true;
                }
            }
            
            double xBias=0.0;
            double yBias=0.0;
            
            if (bRandomTileNoiseStartPoint) {
                UtilityFunctions utility;
                
                xBias=utility.RandomNumberInRange(0.0,1.0)*0.25;
                yBias=utility.RandomNumberInRange(0.0,1.0)*0.25;
            }
                        
            for (int tileJ=tileStartJ;tileJ<tileEndJ;++tileJ) {
                for (int tileI=tileStartI;tileI<tileEndI;++tileI) {
                    
                    // get color
                    double tileX;
                    double tileY;
                    TextureToTileSpace(tileI,tileJ,&tileX,&tileY);
                    
                    if (bRandomTileNoiseStartPoint) {
                        tileX+=xBias*dTiledAreaWidth;
                        tileY+=yBias*dTiledAreaLength;
                    }
                    
                    ColorRGBA colorLocal;
                    
                    if (tileColorDescriptorLocal.colorType=="solid") {
                        colorLocal=colorGlobal;
                    }
                    else if (tileColorDescriptorLocal.colorType=="variable") {
                        if (tileColorDescriptorLocal.colorNoiseType=="perlin"
                            || tileColorDescriptorLocal.colorNoiseType=="cosineModulated"
                            || tileColorDescriptorLocal.colorNoiseType=="wood") {
                            tileColorDescriptorLocal.color=colorGlobal;
                            if (!flipOrientation) {
                                colorLocal=tileColorDescriptorLocal.GetColor(tileX/dTiledAreaWidth,tileY/dTiledAreaLength,&noiseValue);
                            }
                            else {
                                colorLocal=tileColorDescriptorLocal.GetColor(tileY/dTiledAreaLength,tileX/dTiledAreaWidth,&noiseValue);
                            }
                        }
                    }
                    
                    if (tileI>=0 && tileJ>=0 && tileI<iTextureWidth && tileJ<iTextureHeight) {
                        
                        // color texture map
                        textureImage->pixelColor(tileI,tileJ,Color(colorLocal.dR*MaxRGB,colorLocal.dG*MaxRGB,colorLocal.dB*MaxRGB,0));
                        
                        // bump texture value
                        if (!bTileBumpNoiseAlternate) {
                            bumpImage->pixelColor(tileI,tileJ,Color(noiseValue*dMaxTileBumpHeight*MaxRGB,
                                                                    noiseValue*dMaxTileBumpHeight*MaxRGB,
                                                                    noiseValue*dMaxTileBumpHeight*MaxRGB,0));
                        }
                        else {
                            ColorRGBA colorBump;
                            
                            if (!flipOrientation) {
                                colorBump=tileColorDescriptorBump.GetColor(tileX/dTiledAreaWidth,tileY/dTiledAreaLength,&noiseValue);
                            }
                            else {
                                colorBump=tileColorDescriptorBump.GetColor(tileY/dTiledAreaLength,tileX/dTiledAreaWidth,&noiseValue);
                            }
                            
                            bumpImage->pixelColor(tileI,tileJ,Color(noiseValue*dMaxTileBumpHeight*MaxRGB,
                                                                    noiseValue*dMaxTileBumpHeight*MaxRGB,
                                                                    noiseValue*dMaxTileBumpHeight*MaxRGB,0));
                        }
                    }
                }
            }
            
        }
    }
    
}

void RectangularTileMaker::RenderDiscoloration(Image* textureImage)
{

    for (int j=0;j<iTextureHeight;++j) {
        for (int i=0;i<iTextureWidth;++i) {
            
            // tile space coordinates
            double x,y;
            
            TextureToTileSpace(i,j,&x,&y);
            
            // get color
            double noiseValue;
            ColorRGBA color=tileColorDescriptorDiscoloration.GetColor(x,y,&noiseValue);
            
            if (noiseValue>=dDiscolorationThresholdLower && noiseValue<=dDiscolorationThresholdUpper) {
                
                // get current color
                Color pixelColor;
                pixelColor=textureImage->pixelColor(i,j);
                
                ColorRGBA currentColor=ColorRGBA();
                
                currentColor.dR=(double)pixelColor.redQuantum()/MaxRGB;
                currentColor.dG=(double)pixelColor.greenQuantum()/MaxRGB;
                currentColor.dB=(double)pixelColor.blueQuantum()/MaxRGB;
                
                currentColor.dR*=(1.0-dDiscolorationThresholdPercentReduction*0.01);
                currentColor.dG*=(1.0-dDiscolorationThresholdPercentReduction*0.01);
                currentColor.dB*=(1.0-dDiscolorationThresholdPercentReduction*0.01);
                
                //currentColor.dR=1.0;
                //currentColor.dG=0.0;
                //currentColor.dB=0.0;
                
                
                textureImage->pixelColor(i,j,Color(currentColor.dR*MaxRGB,currentColor.dG*MaxRGB,currentColor.dB*MaxRGB,0));
            }
        }
    }

}